#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Comprehensive unit tests for challenge_registry_setup.py"""

# pylint: disable=C0302, C0415, R0904, R0913, R0914, R0915, W0212
import unittest
import sys
import logging
from unittest.mock import Mock, patch, call

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class MockConfig:
    """Mock configuration object for testing"""

    def __init__(self, **kwargs):
        self.email_identifier_support = kwargs.get("email_identifier_support", False)
        self.tnauthlist_support = kwargs.get("tnauthlist_support", False)
        self.dns_persist_01_support = kwargs.get("dns_persist_01_support", False)
        self.forward_address_check = kwargs.get("forward_address_check", False)
        self.reverse_address_check = kwargs.get("reverse_address_check", False)


class TestChallengeRegistrySetup(unittest.TestCase):
    """Test cases for challenge_registry_setup.py functions"""

    def setUp(self):
        """Setup for tests"""
        self.logger = Mock(spec=logging.Logger)
        self.config = MockConfig()

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_001_create_challenge_validator_registry_basic(self):
        """Test basic challenge validator registry creation with minimal config"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_http_instance = Mock()
        mock_dns_instance = Mock()
        mock_tls_instance = Mock()

        mock_http_validator.return_value = mock_http_instance
        mock_dns_validator.return_value = mock_dns_instance
        mock_tls_validator.return_value = mock_tls_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            # Import and test
            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with minimal config (no optional validators)
            config = MockConfig(
                email_identifier_support=False,
                tnauthlist_support=False,
                forward_address_check=False,
                reverse_address_check=False,
            )

            result = create_challenge_validator_registry(self.logger, config)

            # Verify registry creation
            mock_registry.assert_called_once_with(self.logger)
            self.assertEqual(result, mock_registry_instance)

            # Verify standard validators registered
            mock_http_validator.assert_called_once_with(self.logger)
            mock_dns_validator.assert_called_once_with(self.logger)
            mock_tls_validator.assert_called_once_with(self.logger)

            expected_calls = [
                call.register_validator(mock_http_instance),
                call.register_validator(mock_dns_instance),
                call.register_validator(mock_tls_instance),
            ]
            mock_registry_instance.register_validator.assert_has_calls(
                expected_calls, any_order=True
            )

            # Verify optional validators NOT called
            mock_email_validator.assert_not_called()
            mock_tkauth_validator.assert_not_called()
            mock_source_validator.assert_called_once()

            # Verify logging
            self.logger.debug.assert_has_calls(
                [
                    call(
                        "challenge_registry_setup.create_challenge_validator_registry()"
                    ),
                    call(
                        "create_challenge_validator_registry(): Registry created with %d validators: %s",
                        3,
                        "http-01, dns-01, tls-alpn-01",
                    ),
                    call(
                        "challenge_registry_setup.create_challenge_validator_registry() ended"
                    ),
                ]
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_002_create_challenge_validator_registry_email_support(self):
        """Test registry creation with email identifier support enabled"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
            "email-reply-00",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_email_instance = Mock()
        mock_email_validator.return_value = mock_email_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with email support enabled
            config = MockConfig(email_identifier_support=True)

            create_challenge_validator_registry(self.logger, config)

            # Verify email validator registered
            mock_email_validator.assert_called_once_with(self.logger)
            mock_registry_instance.register_validator.assert_any_call(
                mock_email_instance
            )

            # Verify tkauth and source validators NOT called
            mock_tkauth_validator.assert_not_called()
            mock_source_validator.assert_called_once()
            mock_http_validator.assert_called_once()
            mock_dns_validator.assert_called_once()
            mock_tls_validator.assert_called_once()

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_003_create_challenge_validator_registry_all_enabled(self):
        """Test registry creation with all optional features enabled"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
            "email-reply-00",
            "tkauth-01",
            "source-address",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_http_instance = Mock()
        mock_dns_instance = Mock()
        mock_tls_instance = Mock()
        mock_email_instance = Mock()
        mock_tkauth_instance = Mock()
        mock_source_instance = Mock()

        mock_http_validator.return_value = mock_http_instance
        mock_dns_validator.return_value = mock_dns_instance
        mock_tls_validator.return_value = mock_tls_instance
        mock_email_validator.return_value = mock_email_instance
        mock_tkauth_validator.return_value = mock_tkauth_instance
        mock_source_validator.return_value = mock_source_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with all features enabled
            config = MockConfig(
                email_identifier_support=True,
                tnauthlist_support=True,
                forward_address_check=True,
                reverse_address_check=True,
            )

            create_challenge_validator_registry(self.logger, config)

            # Verify all validators created and registered
            mock_http_validator.assert_called_once_with(self.logger)
            mock_dns_validator.assert_called_once_with(self.logger)
            mock_tls_validator.assert_called_once_with(self.logger)

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_003a_create_challenge_validator_registry_dns_persist_support(self):
        """Test registry creation with dns-persist-01 support enabled"""

        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "dns-persist-01",
            "tls-alpn-01",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_dns_persist_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_dns_persist_instance = Mock()
        mock_dns_persist_validator.return_value = mock_dns_persist_instance

        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                DnsPersistChallengeValidator=mock_dns_persist_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                DnsPersistChallengeValidator=mock_dns_persist_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            config = MockConfig(dns_persist_01_support=True)
            create_challenge_validator_registry(self.logger, config)

            mock_dns_persist_validator.assert_called_once_with(self.logger)
            mock_registry_instance.register_validator.assert_any_call(
                mock_dns_persist_instance
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_003b_create_challenge_validator_registry_dns_persist_disabled(self):
        """Test registry creation with dns-persist-01 support disabled."""

        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_dns_persist_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                DnsPersistChallengeValidator=mock_dns_persist_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                DnsPersistChallengeValidator=mock_dns_persist_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            config = MockConfig(dns_persist_01_support=False)
            create_challenge_validator_registry(self.logger, config)

            mock_dns_persist_validator.assert_not_called()

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_004_create_challenge_validator_registry_none_config(self):
        """Test registry creation with None config"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with None config - should handle gracefully
            try:
                create_challenge_validator_registry(self.logger, None)
                # This should raise an AttributeError since None.email_identifier_support would fail
                self.fail("Expected AttributeError for None config")
            except AttributeError:
                # Expected behavior
                pass

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_005_create_challenge_validator_registry_registry_exception(self):
        """Test registry creation when registry constructor raises exception"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry.side_effect = Exception("Registry creation failed")

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            config = MockConfig()

            with self.assertRaises(Exception) as context:
                create_challenge_validator_registry(self.logger, config)

            self.assertEqual(str(context.exception), "Registry creation failed")

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_006_create_custom_registry_basic(self):
        """Test basic custom registry creation"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = ["mock-01", "mock-02"]
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            # Create mock validator classes
            mock_validator_class1 = Mock()
            mock_validator_class2 = Mock()
            mock_validator1 = Mock()
            mock_validator2 = Mock()
            mock_validator_class1.return_value = mock_validator1
            mock_validator_class2.return_value = mock_validator2

            validator_classes = [mock_validator_class1, mock_validator_class2]

            result = create_custom_registry(self.logger, validator_classes)

            # Verify registry creation
            mock_registry.assert_called_once_with(self.logger)
            self.assertEqual(result, mock_registry_instance)

            # Verify validators created and registered
            mock_validator_class1.assert_called_once_with(self.logger)
            mock_validator_class2.assert_called_once_with(self.logger)

            expected_calls = [
                call.register_validator(mock_validator1),
                call.register_validator(mock_validator2),
            ]
            mock_registry_instance.register_validator.assert_has_calls(
                expected_calls, any_order=True
            )

            # Verify logging
            self.logger.info.assert_called_once()
            info_call_args = self.logger.info.call_args[0]
            self.assertIn(
                "Custom challenge validator registry created with 2 validators",
                info_call_args[0] % 2,
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_007_create_custom_registry_empty_validators(self):
        """Test custom registry creation with empty validator list"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = []
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            validator_classes = []

            result = create_custom_registry(self.logger, validator_classes)

            # Verify registry creation
            mock_registry.assert_called_once_with(self.logger)
            self.assertEqual(result, mock_registry_instance)

            # Verify no validators registered
            mock_registry_instance.register_validator.assert_not_called()

            # Verify logging
            self.logger.info.assert_called_once()
            info_call_args = self.logger.info.call_args[0]
            self.assertIn(
                "Custom challenge validator registry created with 0 validators",
                info_call_args[0] % 0,
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_008_create_custom_registry_none_validator_classes(self):
        """Test custom registry creation with None validator classes"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            # Test with None validator classes - should raise TypeError
            with self.assertRaises(TypeError):
                create_custom_registry(self.logger, None)

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_009_create_custom_registry_validator_exception(self):
        """Test custom registry creation when validator constructor raises exception"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            mock_validator_class = Mock()
            mock_validator_class.side_effect = Exception("Validator creation failed")

            validator_classes = [mock_validator_class]

            with self.assertRaises(Exception) as context:
                create_custom_registry(self.logger, validator_classes)

            self.assertEqual(str(context.exception), "Validator creation failed")

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_010_create_custom_registry_registration_exception(self):
        """Test custom registry creation when validator registration raises exception"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.register_validator.side_effect = Exception(
            "Registration failed"
        )
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            mock_validator_class = Mock()
            mock_validator = Mock()
            mock_validator_class.return_value = mock_validator

            validator_classes = [mock_validator_class]

            with self.assertRaises(Exception) as context:
                create_custom_registry(self.logger, validator_classes)

            self.assertEqual(str(context.exception), "Registration failed")

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_011_create_challenge_validator_registry_tnauthlist_support(self):
        """Test registry creation with tnauthlist support enabled"""

        # Mock the module's imports
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
            "tkauth-01",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_tkauth_instance = Mock()
        mock_tkauth_validator.return_value = mock_tkauth_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with tnauthlist support enabled
            config = MockConfig(tnauthlist_support=True)

            create_challenge_validator_registry(self.logger, config)

            # Verify tkauth validator registered
            mock_tkauth_validator.assert_called_once_with(self.logger)
            mock_registry_instance.register_validator.assert_any_call(
                mock_tkauth_instance
            )

            # Verify email and source validators NOT called
            mock_email_validator.assert_not_called()
            mock_source_validator.assert_called_once()
            mock_http_validator.assert_called_once()
            mock_dns_validator.assert_called_once()
            mock_tls_validator.assert_called_once()

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_012_create_challenge_validator_registry_forward_address_check(self):
        """Test registry creation with forward address checking enabled"""

        # Mock the module's imports
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
            "source-address",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_source_instance = Mock()
        mock_source_validator.return_value = mock_source_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with forward address check enabled
            config = MockConfig(forward_address_check=True)

            create_challenge_validator_registry(self.logger, config)

            # Verify source address validator registered with correct parameters
            mock_source_validator.assert_called_once_with(
                self.logger, forward_check=True, reverse_check=False
            )
            mock_registry_instance.register_validator.assert_any_call(
                mock_source_instance
            )

            # Verify email and tkauth validators NOT called
            mock_email_validator.assert_not_called()
            mock_tkauth_validator.assert_not_called()

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_013_create_challenge_validator_registry_reverse_address_check(self):
        """Test registry creation with reverse address checking enabled"""

        # Mock the module's imports
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
            "source-address",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_source_instance = Mock()
        mock_source_validator.return_value = mock_source_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with reverse address check enabled
            config = MockConfig(reverse_address_check=True)

            create_challenge_validator_registry(self.logger, config)

            # Verify source address validator registered with correct parameters
            mock_source_validator.assert_called_once_with(
                self.logger, forward_check=False, reverse_check=True
            )
            mock_registry_instance.register_validator.assert_any_call(
                mock_source_instance
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_014_create_challenge_validator_registry_both_address_checks(self):
        """Test registry creation with both forward and reverse address checking enabled"""

        # Mock the module's imports
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "dns-01",
            "tls-alpn-01",
            "source-address",
        ]
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_source_instance = Mock()
        mock_source_validator.return_value = mock_source_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Test with both address checks enabled
            config = MockConfig(forward_address_check=True, reverse_address_check=True)

            create_challenge_validator_registry(self.logger, config)

            # Verify source address validator registered with both checks
            mock_source_validator.assert_called_once_with(
                self.logger, forward_check=True, reverse_check=True
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_015_create_custom_registry_with_config(self):
        """Test custom registry creation with config parameter"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = ["mock-01"]
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            mock_validator_class = Mock()
            mock_validator = Mock()
            mock_validator_class.return_value = mock_validator

            validator_classes = [mock_validator_class]
            config = {"test": "value"}

            create_custom_registry(self.logger, validator_classes, config)

            # Verify registry creation (config not used in current implementation)
            mock_registry.assert_called_once_with(self.logger)
            mock_validator_class.assert_called_once_with(self.logger)
            mock_registry_instance.register_validator.assert_called_once_with(
                mock_validator
            )

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_016_create_custom_registry_get_supported_types_exception(self):
        """Test custom registry creation when get_supported_types raises exception"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.side_effect = Exception(
            "Get types failed"
        )
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            mock_validator_class = Mock()
            mock_validator = Mock()
            mock_validator_class.return_value = mock_validator

            validator_classes = [mock_validator_class]

            with self.assertRaises(Exception) as context:
                create_custom_registry(self.logger, validator_classes)

            self.assertEqual(str(context.exception), "Get types failed")

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_017_create_custom_registry_mixed_validator_types(self):
        """Test custom registry creation with different validator types"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.get_supported_types.return_value = [
            "http-01",
            "custom-01",
            "test-01",
        ]
        mock_registry.return_value = mock_registry_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import create_custom_registry

            # Create different types of mock validator classes
            class MockHttpValidator:
                def __init__(self, logger):
                    self.logger = logger

            class MockCustomValidator:
                def __init__(self, logger):
                    self.logger = logger

            mock_test_validator_class = Mock()
            mock_test_validator = Mock()
            mock_test_validator_class.return_value = mock_test_validator

            validator_classes = [
                MockHttpValidator,
                MockCustomValidator,
                mock_test_validator_class,
            ]

            result = create_custom_registry(self.logger, validator_classes)

            # Verify registry creation
            self.assertEqual(result, mock_registry_instance)

            # Verify all validator types were handled
            self.assertEqual(mock_registry_instance.register_validator.call_count, 3)

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_018_create_challenge_validator_registry_validator_exception(self):
        """Test registry creation when validator constructor raises exception"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_http_validator.side_effect = Exception("Validator creation failed")
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            config = MockConfig()

            with self.assertRaises(Exception) as context:
                create_challenge_validator_registry(self.logger, config)

            self.assertEqual(str(context.exception), "Validator creation failed")

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_019_create_challenge_validator_registry_missing_config_attributes(self):
        """Test registry creation with config missing some attributes"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            # Create config with missing attributes
            class PartialConfig:
                email_identifier_support = False
                # Missing tnauthlist_support, forward_address_check, reverse_address_check

            config = PartialConfig()

            # Should raise AttributeError when accessing missing attributes
            with self.assertRaises(AttributeError):
                create_challenge_validator_registry(self.logger, config)

    @patch.dict(
        "sys.modules",
        {
            "OpenSSL": Mock(),
            "OpenSSL.crypto": Mock(),
            "acme2certifier.acme_srv.helper": Mock(),
            "acme2certifier.acme_srv.helpers.certificates": Mock(),
            "acme2certifier.acme_srv.challenge_validators": Mock(),
        },
    )
    def test_020_create_challenge_validator_registry_registration_exception(self):
        """Test registry creation when validator registration raises exception"""

        # Mock all the validator classes
        mock_registry = Mock()
        mock_registry_instance = Mock()
        mock_registry_instance.register_validator.side_effect = Exception(
            "Registration failed"
        )
        mock_registry.return_value = mock_registry_instance

        mock_http_validator = Mock()
        mock_dns_validator = Mock()
        mock_tls_validator = Mock()
        mock_email_validator = Mock()
        mock_tkauth_validator = Mock()
        mock_source_validator = Mock()

        mock_http_instance = Mock()
        mock_http_validator.return_value = mock_http_instance

        # Patch both the source module and the target module where imports are used
        with (
            patch.multiple(
                "acme2certifier.acme_srv.challenge_validators",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
            patch.multiple(
                "acme2certifier.acme_srv.challenge_registry_setup",
                ChallengeValidatorRegistry=mock_registry,
                HttpChallengeValidator=mock_http_validator,
                DnsChallengeValidator=mock_dns_validator,
                TlsAlpnChallengeValidator=mock_tls_validator,
                EmailReplyChallengeValidator=mock_email_validator,
                TkauthChallengeValidator=mock_tkauth_validator,
                SourceAddressValidator=mock_source_validator,
            ),
        ):

            from acme2certifier.acme_srv.challenge_registry_setup import (
                create_challenge_validator_registry,
            )

            config = MockConfig()

            with self.assertRaises(Exception) as context:
                create_challenge_validator_registry(self.logger, config)

            self.assertEqual(str(context.exception), "Registration failed")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""unittests for acme2certifier"""

# pylint: disable= C0415, W0212
import unittest
import sys
import os
from unittest.mock import patch, Mock, mock_open
import base64
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend
from cryptography import x509

sys.path.insert(0, ".")
sys.path.insert(1, "..")


class TestACMEHandler(unittest.TestCase):
    """test class for cgi_handler"""

    def setUp(self):
        """setup unittest"""
        import logging

        logging.basicConfig(level=logging.CRITICAL)
        self.logger = logging.getLogger("test_a2c")
        from acme2certifier.cahandlers.pkcs7_soap_ca_handler import (
            CAhandler,
            binary_read,
            binary_write,
        )

        self.cahandler = CAhandler(False, self.logger)
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.binary_read = binary_read
        self.binary_write = binary_write

    def test_001_default(self):
        """default test which always passes"""
        self.assertEqual("foo", "foo")

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_002_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        mock_load_cfg.return_value = {}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        self.assertIn(
            "ERROR:test_a2c:CAhandler section is missing in configuration file.",
            lcm.output,
        )

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_003_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        mock_load_cfg.return_value = {"CAhandler": {"foo": "bar"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_004_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        mock_load_cfg.return_value = {"CAhandler": {"email": "email", "foo": "bar"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertEqual("email", self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_005_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        mock_load_cfg.return_value = {"CAhandler": {"soap_srv": "soap_srv"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertEqual("soap_srv", self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_006_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        mock_load_cfg.return_value = {"CAhandler": {"profilename": "profilename"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertEqual("profilename", self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_007_config_load(self, mock_load_cfg):
        """test _config_load no cahandler section"""
        mock_load_cfg.return_value = {"CAhandler": {"ca_bundle": "ca_bundle"}}

        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertEqual("ca_bundle", self.cahandler.ca_bundle)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_008_config_load(self, mock_load_cfg, mock_file):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = False
        mock_load_cfg.return_value = {"CAhandler": {"signing_cert": "signing_cert"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate file not found: signing_cert",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.x509.load_pem_x509_certificate")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_009_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_cert"
        mock_load_cfg.return_value = {"CAhandler": {"signing_cert": "signing_cert"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertEqual("signing_cert", self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing key option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_010_config_load(self, mock_load_cfg, mock_file):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = False
        mock_load_cfg.return_value = {"CAhandler": {"password": "password"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertEqual("password".encode("utf8"), self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)

    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_011_config_load(self, mock_load_cfg, mock_file):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = False
        mock_load_cfg.return_value = {"CAhandler": {"signing_key": "signing_key"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
            "ERROR:test_a2c:Signing key file not found: signing_key",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_012_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {"CAhandler": {"signing_key": "signing_key"}}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertEqual("signing_key", self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertFalse(self.cahandler.signing_script_dic)
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:Signing certificate option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_013_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {"CAhandler": {"signing_script": "signing_script"}}
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {"signing_script": "signing_script"}, self.cahandler.signing_script_dic
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_alias option is missing in configuration file.",
            "ERROR:test_a2c:signing_csr_path option is missing in configuration file.",
            "ERROR:test_a2c:signing_config_variant option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_014_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {
            "CAhandler": {
                "signing_script": "signing_script",
                "signing_alias": "signing_alias",
            }
        }
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {"signing_alias": "signing_alias", "signing_script": "signing_script"},
            self.cahandler.signing_script_dic,
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_csr_path option is missing in configuration file.",
            "ERROR:test_a2c:signing_config_variant option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_015_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {
            "CAhandler": {
                "signing_script": "signing_script",
                "signing_csr_path": "signing_csr_path",
            }
        }
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {
                "signing_csr_path": "signing_csr_path",
                "signing_script": "signing_script",
            },
            self.cahandler.signing_script_dic,
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_alias option is missing in configuration file.",
            "ERROR:test_a2c:signing_config_variant option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_016_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {
            "CAhandler": {
                "signing_script": "signing_script",
                "signing_config_variant": "signing_config_variant",
            }
        }
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {
                "signing_script": "signing_script",
                "signing_config_variant": "signing_config_variant",
            },
            self.cahandler.signing_script_dic,
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_alias option is missing in configuration file.",
            "ERROR:test_a2c:signing_csr_path option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_017_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {
            "CAhandler": {
                "signing_script": "signing_script",
                "signing_user": "signing_user",
            }
        }
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {"signing_script": "signing_script", "signing_user": "signing_user"},
            self.cahandler.signing_script_dic,
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_alias option is missing in configuration file.",
            "ERROR:test_a2c:signing_csr_path option is missing in configuration file.",
            "ERROR:test_a2c:signing_config_variant option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_018_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {
            "CAhandler": {
                "signing_script": "signing_script",
                "signing_sleep_timer": "signing_sleep_timer",
            }
        }
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {
                "signing_script": "signing_script",
                "signing_sleep_timer": "signing_sleep_timer",
            },
            self.cahandler.signing_script_dic,
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_alias option is missing in configuration file.",
            "ERROR:test_a2c:signing_csr_path option is missing in configuration file.",
            "ERROR:test_a2c:signing_config_variant option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    @patch("cryptography.hazmat.primitives.serialization.load_pem_private_key")
    @patch("os.path.exists")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.load_config")
    def test_019_config_load(self, mock_load_cfg, mock_file, mock_load):
        """test _config_load signing cert configured but does not exist"""
        mock_file.return_value = True
        mock_load.return_value = "signing_key"
        mock_load_cfg.return_value = {
            "CAhandler": {
                "signing_script": "signing_script",
                "signing_interpreter": "signing_interpreter",
            }
        }
        self.maxDiff = None
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.cahandler._config_load()
        self.assertFalse(self.cahandler.soap_srv)
        self.assertFalse(self.cahandler.profilename)
        self.assertFalse(self.cahandler.email)
        self.assertFalse(self.cahandler.signing_cert)
        self.assertFalse(self.cahandler.signing_key)
        self.assertFalse(self.cahandler.password)
        self.assertEqual(
            {
                "signing_script": "signing_script",
                "signing_interpreter": "signing_interpreter",
            },
            self.cahandler.signing_script_dic,
        )
        error_buffer = [
            "ERROR:test_a2c:SOAP server URL (soap_srv) is missing in configuration file.",
            "WARNING:test_a2c:SOAP server certificate validation is disabled.",
            "ERROR:test_a2c:Profile name (profilename) is missing in configuration file.",
            "ERROR:test_a2c:Email option is missing in configuration file.",
            "ERROR:test_a2c:signing_alias option is missing in configuration file.",
            "ERROR:test_a2c:signing_csr_path option is missing in configuration file.",
            "ERROR:test_a2c:signing_config_variant option is missing in configuration file.",
        ]
        self.assertEqual(error_buffer, lcm.output)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._config_load")
    def test_020_enter(self, mock_cfgload):
        """enter - no soap server configured"""
        self.cahandler.__enter__()
        self.assertTrue(mock_cfgload.called)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._config_load")
    def test_021_enter(self, mock_cfgload):
        """enter soap server configured"""
        self.cahandler.soap_srv = "mock_srv"
        self.cahandler.__enter__()
        self.assertFalse(mock_cfgload.called)

    def test_022_exit(self):
        """enter - no soap server configured"""
        self.cahandler.__exit__()

    @patch("pyasn1.codec.der.decoder.decode")
    def test_023_cert_decode(self, mock_der):
        """test _cert_decode()"""
        mock_der.return_value = "decode"
        cert = Mock()
        cert.public_bytes = Mock()
        self.assertEqual("decode", self.cahandler._cert_decode(cert))

    def test_024_poll(self):
        """test poll"""
        self.assertEqual(
            (None, None, None, "poll_identifier", False),
            self.cahandler.poll("cert_name", "poll_identifier", "csr"),
        )

    def test_025_revoke(self):
        """test revoke"""
        self.assertEqual(
            (
                500,
                "urn:ietf:params:acme:error:serverInternal",
                "Revocation is not supported.",
            ),
            self.cahandler.revoke("cert_name", "reason", "date"),
        )

    def test_026_trigger(self):
        """test revoke"""
        self.assertEqual((None, None, None), self.cahandler.trigger("identifier"))

    def test_027_soaprequest_build(self):
        """test soap request build"""
        self.cahandler.profilename = "profilename"
        self.cahandler.email = "email"
        pkcs7 = "pkcs7"
        result = """
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:aur="http://monetplus.cz/services/kb/aurora">
<soapenv:Header/>
<soapenv:Body>
    <aur:RequestCertificate>
        <aur:request>
            <aur:ProfileName>profilename</aur:ProfileName>
            <aur:CertificateRequestRaw>pkcs7</aur:CertificateRequestRaw>
            <aur:Email>email</aur:Email>
            <aur:ReturnCertificateCaChain>true</aur:ReturnCertificateCaChain>
        </aur:request>
    </aur:RequestCertificate>
</soapenv:Body>
</soapenv:Envelope>\n"""
        self.assertEqual(result, self.cahandler._soaprequest_build(pkcs7))

    @patch("os.path.isfile")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_028_binary_read(self, mock_isfile):
        """test read binary file"""
        mock_isfile.return_value = True
        self.assertEqual("foo", self.binary_read(self.logger, "filename"))

    @patch("os.path.isdir")
    @patch("builtins.open", mock_open(read_data="foo"), create=True)
    def test_029_binary_write(self, mock_isdir):
        """test wrote binary file"""
        mock_isdir.return_value = True
        self.assertFalse(self.binary_write(self.logger, "filename", "content"))

    def test_030_sign(self):
        """test _sign unkown key format"""
        key = "key"
        payload = "foo"
        self.assertEqual((None, None), self.cahandler._sign(key, payload))

    @patch("cryptography.hazmat.primitives.asymmetric.rsa")
    def test_031_sign(self, mock_rsa):
        """test _sign rsa key"""
        keyph = b"Test1234"
        with open(self.dir_path + "/ca/sub-ca-key.pem", "rb") as open_file:
            key = serialization.load_pem_private_key(
                open_file.read(), password=keyph, backend=default_backend()
            )
        payload = b"foo"
        result = self.cahandler._sign(key, payload)
        signature = b"4oTEIybGnmkfnG+Fvf0t8Sx8YHSf55tm3WtcdPagvtNM3vLjsidWKc2yliGYVmDqT9E+/wx3tvsMeDrgRiAzMhbjPYOeKwyx30BZT++4Fw9OkRQyriwyLB3ncFReVF8DyBRj/3S1Ftoy6Msa2CCk59LhYm/ubBQAm88gYiBzCFtVhneNOg5vS2s79UuyLjE2J90Yjs3z7OCckWrZ1UxI3UBoaJAWQg83M6fnF4aMkpnO3Jd6oQ4nq7r4EeVKYYEwrOINKKfh/1ykaCLg2K9OAD2LY1b9LilHTG8lcoUhS+bBMJkESHi508EzFQ4IUdsA42porTkEkdc5g9ZmCm7PPjroSRZGtM00R6aV/4z8Tlp4JBaov9x3fUd5wKjGIP0mdLQamAfxhK/pUqzM/lXtndprV7yh07tzypHa1XNvmTn/di2jNu90cq3eGgi3nBY98u+GcHTFnFH2aW2hk7kxqmxT4ymsZhlviIX8GIT4blE2nJgcl91Ktxm9QataRMjny/uJd//olQAXGMcbDwhNpYBfdJe99XoeuY+xNtJtlQt7IciTmJ3DEcK2kTtsNZ2i/lvn+iYR4iD9fJ/S4FedHqPZi48Q+LSnGC61zD21ZgbT8FrzUTnmmgw9BeDTWezGDGgBdOIuG313waZlvdDahk+6AYz9tOxS+bm9Epcj3NY="
        alg = """AlgorithmIdentifier:\n algorithm=1.2.840.113549.1.1.11\n parameters=0x0500\n"""
        self.assertEqual(signature, base64.b64encode(result[0]))
        self.assertEqual(alg, str(result[1]))

    @patch("cryptography.hazmat.primitives.asymmetric.rsa")
    def test_032_sign(self, mock_rsa):
        """test _sign ecc key"""
        ecc_key = b"-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEIGCu1fYGkqMdPtsNH7xVc8QBjCWCkcUTVKX6f8vLhtkvoAoGCCqGSM49\nAwEHoUQDQgAEan72++swi7J5B1HVYp1CjXPqckkQquiMIQhz5xYesv9f4KK/ouKS\n1uJ3ZYwPbWUsDd8/03vf9VdlfZzL3W3ZQw==\n-----END EC PRIVATE KEY-----"
        key = serialization.load_pem_private_key(
            ecc_key, password=None, backend=default_backend()
        )
        payload = b"foo"
        result = self.cahandler._sign(key, payload)
        alg = """AlgorithmIdentifier:\n algorithm=1.2.840.10045.4.3.2\n"""
        self.assertEqual(alg, str(result[1]))

    def test_033_certraw_get(self):
        """ test _certraw_get """ ""
        with open(self.dir_path + "/ca/sub-ca-client.pem", "r") as fso:
            pem_data = fso.read()
        result = "MIIEGDCCAgCgAwIBAgIJALL8aztMPfV2MA0GCSqGSIb3DQEBCwUAMEgxCzAJBgNVBAYTAkRFMQ8wDQYDVQQIDAZCZXJsaW4xFzAVBgNVBAoMDkFjbWUyQ2VydGlmaWVyMQ8wDQYDVQQDDAZzdWItY2EwHhcNMTkwNjI1MDEyNTAwWhcNMjAwNjI1MDEyNTAwWjBPMQswCQYDVQQGEwJERTEPMA0GA1UEBxMGQmVybGluMRcwFQYDVQQKEw5BY21lMkNlcnRpZmllcjEWMBQGA1UEAwwNY2xpZW50X3N1Yi1jYTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALvoKKg3ciBVWZtquiWyMogWU6ydEfmLbXktK6T+owxzxHVaoePVGH9DZvTZD2pHS8xJ6fpFr3pZYiuqiUHuxdMpj9gVxik5ivBrSJIkZXLxwvNJWpMa1o1Hxz1By3Hrlm3ebKIzfQPqRRcdjWtJgCFbcTpalwhE1RQFMp4Icb08aAE9uEaZQ4uZ8Ls30J6IHC4PG63lGI1tkAtLIoUWupRAmnWDx0ysXzXeN7m+Lff9ols9MZNgzRMgY/zGUq0LzZfi+L+Iev3sztCdoIOBA/K63jv0hOPyYg331L05XIwbLeUoUG41J4pZzafx6MAFp4Zam1w+aafCzEw7ZPHQvn0CAwEAATANBgkqhkiG9w0BAQsFAAOCAgEABPgWo4KAXJNXNfEBbixDuCxtwO1JuphSOTcpIlEp+uNOSDzgNEbrhUXTNM8SPshzFjBpudc29okiyC62CfLD/X+EvIeKo/oa477kN6MuNfqLGZ42a935ES3S00Wy8rbwyIoPCsKWT/6VsHRHUn8XhFNFUBKZ8FGxwXcAVpPanyikURqVH1MgAk62hJQdYjSxdga/GKS1dS39fyxQz7uBPt5WIQZPzL6dr2Yn/4lQUvTUVus2e1cTh3z02yB5EDlEAcMMvMNpfYvNdU5H6QEPwysbkW9E/Ep84aq21zwuPxICh0KdjHWKkHtCqDoEYIADDl1AD5UdJTMQ9LIzUjsBvtB5I6yT7jgsx/iqTDrkJVK/zRf4NeKRa3AW57jsPUIcUstUFnVJbg+MM4fYmapx8Hqm/Aq+II9ip80AM6hXvierTQn4MNQivL0ZJfj0Ro9KEIDAHN3IAfIlFovbkBPLMi9PtfyhuVmXpthE9OaDlgUguWb45LAKwgfu1TFGPPpf5jTw2qVx0F+iCiUwK8ZgnakkXOKE5+KIb8ejL+3pPd5Wt+45w/7gEFOjT6XAzZGnUtcMH/lpxmgbl3/SKkyrW4h7PnF2FEEVC4XnZuQm+ZwD/PpXfmAA52ygKHBzUr9V33CkW0FhvjqkAUya5x9CqWlHoal0RVvFavnw+4ImqbE="
        self.assertEqual(result, self.cahandler._certraw_get(pem_data))

    def test_034_pkcs7_create(self):
        """test pkcs7_create"""
        keyph = b"Test1234"
        with open(self.dir_path + "/ca/csr.der", "rb") as open_file:
            csr_der = open_file.read()
        with open(self.dir_path + "/ca/sub-ca-key.pem", "rb") as open_file:
            signing_key = serialization.load_pem_private_key(
                open_file.read(), password=keyph, backend=default_backend()
            )
        with open(self.dir_path + "/ca/sub-ca-cert.pem", "rb") as open_file:
            signing_cert = x509.load_pem_x509_certificate(
                open_file.read(), default_backend()
            )

        decoded_cert = self.cahandler._cert_decode(signing_cert)
        expected_result = b"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"
        _error, result = self.cahandler._pkcs7_create(
            decoded_cert, csr_der, signing_key
        )
        self.assertEqual(expected_result, base64.b64encode(result))

    @patch("requests.post")
    def test_035_soaprequest_send(self, mock_post):
        """soaprequest_send() - request exception"""
        mock_post.side_effect = Exception("exc_api_post")
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("Connection error", None), self.cahandler._soaprequest_send("payload")
            )
        self.assertIn(
            "ERROR:test_a2c:SOAP request to CA failed: exc_api_post", lcm.output
        )

    @patch("xmltodict.parse")
    @patch("requests.post")
    def test_036_soaprequest_send(self, mock_post, mock_xml_parse):
        """soaprequest_send() - 200 xml-parsing error"""
        mock_post.return_value = Mock(status_code=200)
        mock_xml_parse.return_value = {"foo": "bar"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("Parsing error", None), self.cahandler._soaprequest_send("payload")
            )
        self.assertIn(
            "ERROR:test_a2c:XML parsing error in SOAP response from CA.",
            lcm.output,
        )

    @patch("xmltodict.parse")
    @patch("requests.post")
    def test_037_soaprequest_send(self, mock_post, mock_xml_parse):
        """soaprequest_send() - 200 xml-parsing successful"""
        mock_post.return_value = Mock(status_code=200)
        mock_xml_parse.return_value = {
            "s:Envelope": {
                "s:Body": {
                    "RequestCertificateResponse": {
                        "RequestCertificateResult": {"IssuedCertificate": "foo"}
                    }
                }
            }
        }
        self.assertEqual((None, "foo"), self.cahandler._soaprequest_send("payload"))

    @patch("xmltodict.parse")
    @patch("requests.post")
    def test_038_soaprequest_send(self, mock_post, mock_xml_parse):
        """soaprequest_send() - 400 xml-parsing error"""
        mock_post.return_value = Mock(status_code=400)
        mock_xml_parse.return_value = {"foo": "bar"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("Server error", None), self.cahandler._soaprequest_send("payload")
            )
        self.assertIn(
            "ERROR:test_a2c:CA server returned HTTP error status: 400",
            lcm.output,
        )
        self.assertIn(
            "ERROR:test_a2c:Unknown error while parsing SOAP response from CA.",
            lcm.output,
        )

    @patch("xmltodict.parse")
    @patch("requests.post")
    def test_039_soaprequest_send(self, mock_post, mock_xml_parse):
        """soaprequest_send() - 400 xml-parsing successful"""
        mock_post.return_value = Mock(status_code=400)
        mock_xml_parse.return_value = {
            "s:Envelope": {
                "s:Body": {
                    "s:Fault": {"faultcode": "faultcode", "faultstring": "faultstring"}
                }
            }
        }
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("Server error", None), self.cahandler._soaprequest_send("payload")
            )
        self.assertIn(
            "ERROR:test_a2c:CA server returned HTTP error status: 400",
            lcm.output,
        )
        self.assertIn(
            "ERROR:test_a2c:SOAP response contains faultcode: faultcode",
            lcm.output,
        )
        self.assertIn(
            "ERROR:test_a2c:SOAP response contains faultstring: faultstring",
            lcm.output,
        )

    def test_040_get_certificates(self):
        """test pkcs7_create"""
        with open(self.dir_path + "/ca/certs_der.p7b", "rb") as open_file:
            pkcs7_bundle = open_file.read()
        result = [
            "-----BEGIN CERTIFICATE-----\nMIIFTzCCAzegAwIBAgIIAzHyhSyrXfMwDQYJKoZIhvcNAQELBQAwKzEXMBUGA1UE\nCxMOYWNtZTJjZXJ0aWZpZXIxEDAOBgNVBAMTB3Jvb3QtY2EwHhcNMjAwNTI3MTM1\nNDAwWhcNMzAwNTI2MjM1OTAwWjAqMRcwFQYDVQQLEw5hY21lMmNlcnRpZmllcjEP\nMA0GA1UEAxMGc3ViLWNhMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA\nxXHaGZsolXe+PBdUryngHP9VbBC1mehqeTtYI+hqsqGNH7q9a7bSrxMwFuF1kYL8\njqqxkJdtl0L94xcxJg/ZdMx7Nt0vGI+BaAuTpEpUEHeN4tqS6NhB/m/0LGkAELc/\nqkzmoO4B1FDwEEj/3IXtZcupqG80oDt7jWSGXdtF7NTjzcumznMeRXidCdhxRxT/\n/WrsChaytXo0xWZ56oeNwd6x6Dr8/39PBOWtj4fldyDcg+Q+alci2tx9pxmu2bCV\nXcB9ftCLKhDk2WEHE88bgKSp7fV2RCmq9po+Tx8JJ7qecLunUsK/F0XN4kpoQLm9\nhcymqchnMSncSiyin1dQHGHWgXDtBDdq6A2Z6rx26Qk5H9HTYvcNSe1YwFEDoGLB\nZQjbCPWiaqoaH4agBQTclPvrrSCRaVmhUSO+pBtSXDkmN4t3MDZxfgRkp8ixwkB1\n5Y5f0LTpCyAJsdQDw8+Ea0aDqO30eskh4CErnm9+Fejd9Ew2cwpdwfBXzVSbYilM\nGueQihZHvJmVRxAwU69aO2Qs8B0tQ60CfWKVlmWPiakrvYYlPp0FBsM61G6LZEN8\nhH2CKnS8hHv5IWEXZvp0Pk8V3P5h6bWN0Tl+x/V1Prt7Wp8NoiPETE8XyDDxe6dm\nKxztWBH/mTsJyMGb6ZiUoXdPU9TFUKqHxTRLHaxfsPsCAwEAAaN4MHYwEgYDVR0T\nAQH/BAgwBgEB/wIBATAdBgNVHQ4EFgQUv96OjgYiIqutQ8jd1E+oq0hBPtUwDgYD\nVR0PAQH/BAQDAgGGMBEGCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYP\neGNhIGNlcnRpZmljYXRlMA0GCSqGSIb3DQEBCwUAA4ICAQBbHLEVyg4f9uEujroc\n31UVyDRLMdPgEPLjOenSBCBmH0N81whDmxNI/7JAAB6J14WMX8OLF0HkZnb7G77W\nvDhy1aFvQFbXHBz3/zUO9Mw9J4L2XEW6ond3Nsh1m2oXeBde3R3ANxuIzHqZDlP9\n6YrRcHjnf4+1/5AKDJAvJD+gFb5YnYUKH2iSvHUvG17xcZx98Rf2eo8LealG4JqH\nJh4sKRy0VjDQD7jXSCbweTHEb8wz+6OfNGrIo+BhTFP5vPcwE4nlJwYBoaOJ5cVa\n7gdQJ7WkLSxvwHxuxzvSVK73u3jl3I9SqTrbMLG/jeJyV0P8EvdljOaGnCtQVRwC\nzM4ptXUvKhKOHy7/nyTF/Bc35ZwwL/2xWvNK1+NibgE/6CFxupwWpdmxQbVVuoQ3\n2tUil9ty0yC6m5GKE8+t1lrZuxyA+b/TBnYNO5xo8UEMbkpxaNYSwmw+f/loxXP/\nM7sIBcLvy2ugHEBxwd9o/kLXeXT2DaRvxPjp4yk8MpJRpNmz3aB5HJwaUnaRLVo5\nZ3XWWXmjMGZ6/m0AAoDbDz/pXtOoJZT8BJdD1DuDdszVsQnLVn4B/LtIXL6FbXsF\nzfv6ERP9a5gpKUZ+4NjgrnlGtdccNZpwyWF0IXcvaq3b8hXIRO4hMjzHeHfzJN4t\njX1vlY35Ofonc4+6dRVamBiF9A==\n-----END CERTIFICATE-----\n",
            "-----BEGIN CERTIFICATE-----\nMIIFcDCCA1igAwIBAgIIevLTTxOMoZgwDQYJKoZIhvcNAQELBQAwKzEXMBUGA1UE\nCxMOYWNtZTJjZXJ0aWZpZXIxEDAOBgNVBAMTB3Jvb3QtY2EwHhcNMjAwNTI3MDAw\nMDAwWhcNMzAwNTI2MjM1OTU5WjArMRcwFQYDVQQLEw5hY21lMmNlcnRpZmllcjEQ\nMA4GA1UEAxMHcm9vdC1jYTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB\nAJy4UZHdZgYt64k/rFamoC676tYvtabeuiqVw1c6oVZI897cFLG6BYwyr2Eaj7tF\nrqTJDeMN4vZSudLsmLDq6m8KwX/riPzUTIlcjM5aIMANZr9rLEs3NWtcivolB5aQ\n1slhdVitUPLuxsFnYeQTyxFyP7lng9M/Z403KLG8phdmKjM0vJkaj4OuKOXf3UsW\nqWQYyRl/ms07xVj02uq08LkoeO+jtQisvyVXURdaCceZtyK/ZBQ7NFCsbK112cVR\n1e2aJol7NJAA6Wm6iBzAdkAA2l3kh40SLoEbaiaVMixLN2vilIZOOAoDXX4+T6ir\n+KnDVSJ2yu5c/OJMwuXwHrh7Lgg1vsFR5TNehknhjUuWOUO+0TkKPg2A7KTg72OZ\n2mOcLZIbxzr1P5RRvdmLQLPrTF2EJvpQPNmbXqN3ZVWEvfHTjkkTFY/dsOTvFTgS\nri15zYKch8votcU7z+BQhgmMtwO2JhPMmZ6ABd9skI7ijWpwOltAhxtdoBO6T6CB\nCrE2yXc6V/PyyAKcFglNmIght5oXsnE+ub/dtx8f9Iea/xNPdo5aGy8fdaitolDK\n16kd3Kb7OE4HMHIwOxxF1BEAqerxxhbLMRBr8hRSZI5cvLzWLvpAQ5zuhjD6V3b9\nBYFd4ujAu3zl3mbzdbYjFoGOX6aBZaGDxlc4O2W7HxntAgMBAAGjgZcwgZQwDwYD\nVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQUDGVvuTFYZtEAkz3af9wRKDDvAswwHwYD\nVR0jBBgwFoAUDGVvuTFYZtEAkz3af9wRKDDvAswwDgYDVR0PAQH/BAQDAgGGMBEG\nCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYPeGNhIGNlcnRpZmljYXRl\nMA0GCSqGSIb3DQEBCwUAA4ICAQAjko7dX+iCgT+m3Iy1Vg6j7MRevPAzq1lqHRRN\nNdt2ct530pIut7Fv5V2xYk35ka+i/G+XyOvTXa9vAUKiBtiRnUPsXu4UcS7CcrCX\nEzHx4eOtHnp5wDhO0Fx5/OUZTaP+L7Pd1GD/j953ibx5bMa/M9Rj+S486nst57tu\nDRmEAavFDiMd6L3jH4YSckjmIH2uSeDIaRa9k6ag077XmWhvVYQ9tuR7RGbSuuV3\nFc6pqcFbbWpoLhNRcFc+hbUKOsKl2cP+QEKP/H2s3WMllqgAKKZeO+1KOsGo1CDs\n475bIXyCBpFbH2HOPatmu3yZRQ9fj9ta9EW46n33DFRNLinFWa4WJs4yLVP1juge\n2TCOyA1t61iy++RRXSG3e7NFYrEZuCht1EdDAdzIUY89m9NCPwoDYS4CahgnfkkO\n7YQe6f6yqK6isyf8ZFcp1uF58eERDiF/FDqS8nLmCdURuI56DDoNvDpig5J/9RNW\nG8vEvt2p7QrjeZ3EAatx5JuYty/NKTHZwJWk51CgzEgzDwzE2JIiqeldtL5d0Sl6\neVuv0G04BEyuXxEWpgVVzBS4qEFIBSnTJzgu1PXmId3yLvg2Nr8NKvwyZmN5xKFp\n0A9BWo15zW1PXDaD+l39oTYD7agjXkzTAjYIcfNJ7ATIYFD0xAvNAOf70s7aNupF\nfvkG2Q==\n-----END CERTIFICATE-----\n",
        ]
        self.assertEqual(result, self.cahandler._get_certificate(pkcs7_bundle))

    def test_041_pkcs7_signing_config_verify(self):
        """test _pkcs7_signing_config_verify()"""
        self.cahandler.signing_script_dic = {}
        self.assertEqual(
            "signing config incomplete: option signing_script is missing",
            self.cahandler._pkcs7_signing_config_verify(),
        )

    def test_042_pkcs7_signing_config_verify(self):
        """test _pkcs7_signing_config_verify()"""
        self.cahandler.signing_script_dic = {"signing_script": "signing_script"}
        self.assertEqual(
            "signing config incomplete: option signing_alias is missing",
            self.cahandler._pkcs7_signing_config_verify(),
        )

    def test_043_pkcs7_signing_config_verify(self):
        """test _pkcs7_signing_config_verify()"""
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_alias": "signing_alias",
        }
        self.assertEqual(
            "signing config incomplete: option signing_csr_path is missing",
            self.cahandler._pkcs7_signing_config_verify(),
        )

    @patch("os.path.isdir")
    def test_044_pkcs7_signing_config_verify(self, mock_path):
        """test _pkcs7_signing_config_verify()"""
        mock_path.return_value = False
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_alias": "signing_alias",
            "signing_csr_path": "signing_csr_path",
        }
        self.assertEqual(
            "signing_csr_path signing_csr_path does not exist or is not a directory",
            self.cahandler._pkcs7_signing_config_verify(),
        )

    @patch("os.path.isdir")
    def test_045_pkcs7_signing_config_verify(self, mock_path):
        """test _pkcs7_signing_config_verify()"""
        mock_path.return_value = True
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_alias": "signing_alias",
            "signing_csr_path": "signing_csr_path",
        }
        self.assertEqual(
            "signing config incomplete: option signing_config_variant is missing",
            self.cahandler._pkcs7_signing_config_verify(),
        )

    @patch("os.path.isdir")
    def test_046_pkcs7_signing_config_verify(self, mock_path):
        """test _pkcs7_signing_config_verify()"""
        mock_path.return_value = True
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_alias": "signing_alias",
            "signing_csr_path": "signing_csr_path",
            "signing_config_variant": "signing_config_variant",
        }
        self.assertEqual(None, self.cahandler._pkcs7_signing_config_verify())

    def test_047_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {}
        self.assertEqual(
            [], self.cahandler._signing_command_build("csr_unsigned", "csr_signed")
        )

    def test_048_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {}
        self.assertEqual(
            [], self.cahandler._signing_command_build("csr_unsigned", "csr_signed")
        )

    def test_049_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {"signing_user": "signing_user"}
        self.assertEqual(
            [], self.cahandler._signing_command_build("csr_unsigned", "csr_signed")
        )

    def test_050_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {
            "signing_user": "signing_user",
            "signing_script": "signing_script",
        }
        self.assertEqual(
            ["sudo", "signing_user", "signing_script", "csr_unsigned", "csr_signed"],
            self.cahandler._signing_command_build("csr_unsigned", "csr_signed"),
        )

    def test_051_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {
            "signing_user": "signing_user",
            "signing_script": "signing_script",
            "signing_interpreter": "signing_interpreter",
        }
        self.assertEqual(
            [
                "sudo",
                "signing_user",
                "signing_interpreter",
                "signing_script",
                "csr_unsigned",
                "csr_signed",
            ],
            self.cahandler._signing_command_build("csr_unsigned", "csr_signed"),
        )

    def test_052_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_interpreter": "signing_interpreter",
        }
        self.assertEqual(
            ["signing_interpreter", "signing_script", "csr_unsigned", "csr_signed"],
            self.cahandler._signing_command_build("csr_unsigned", "csr_signed"),
        )

    def test_053_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_interpreter": "signing_interpreter",
            "signing_alias": "signing_alias",
        }
        self.assertEqual(
            ["signing_interpreter", "signing_script", "csr_unsigned", "csr_signed"],
            self.cahandler._signing_command_build("csr_unsigned", "csr_signed"),
        )

    def test_054_signing_command_build(self):
        """test _signing_command_build()"""
        self.cahandler.signing_script_dic = {
            "signing_script": "signing_script",
            "signing_interpreter": "signing_interpreter",
            "signing_alias": "signing_alias",
            "signing_config_variant": "signing_config_variant",
        }
        self.assertEqual(
            [
                "signing_interpreter",
                "signing_script",
                "csr_unsigned",
                "csr_signed",
                "signing_alias",
                "signing_config_variant",
            ],
            self.cahandler._signing_command_build("csr_unsigned", "csr_signed"),
        )

    @patch("os.remove")
    @patch("os.path.isfile")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_read")
    @patch("subprocess.call")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_write")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._signing_command_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.generate_random_string")
    @patch(
        "acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_signing_config_verify"
    )
    def test_055_pkcs7_sign_external(
        self,
        mock_vrf,
        mock_rand,
        mock_build,
        mock_write,
        mock_call,
        mock_read,
        mock_file,
        mock_rm,
    ):
        """test _pkcs7_sign_external()"""
        mock_vrf.return_value = True
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(
                ("Config incomplete", None), self.cahandler._pkcs7_sign_external("csr")
            )
        self.assertIn(
            "ERROR:test_a2c:External signing configuration is incomplete: True",
            lcm.output,
        )
        self.assertFalse(mock_rand.called)
        self.assertFalse(mock_build.called)
        self.assertFalse(mock_write.called)
        self.assertFalse(mock_call.called)
        self.assertFalse(mock_read.called)
        self.assertFalse(mock_file.called)
        self.assertFalse(mock_rm.called)

    @patch("os.remove")
    @patch("os.path.isfile")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_read")
    @patch("subprocess.call")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_write")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._signing_command_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.generate_random_string")
    @patch(
        "acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_signing_config_verify"
    )
    def test_056_pkcs7_sign_external(
        self,
        mock_vrf,
        mock_rand,
        mock_build,
        mock_write,
        mock_call,
        mock_read,
        mock_file,
        mock_rm,
    ):
        """test _pkcs7_sign_external() all good"""
        mock_vrf.return_value = False
        mock_read.return_value = "foo"
        mock_call.return_value = None
        mock_file.return_value = True
        self.cahandler.signing_script_dic = {"signing_csr_path": "signing_csr_path"}
        # with self.assertLogs('test_a2c', level='INFO') as lcm:
        self.assertEqual((None, "foo"), self.cahandler._pkcs7_sign_external("csr"))
        # self.assertIn('ERROR:test_a2c:CAhandler._pkcs7_sign_external(): config incomplete: True', lcm.output)
        self.assertTrue(mock_rand.called)
        self.assertTrue(mock_build.called)
        self.assertTrue(mock_write.called)
        self.assertTrue(mock_call.called)
        self.assertTrue(mock_read.called)
        self.assertTrue(mock_file.called)
        self.assertTrue(mock_rm.called)

    @patch("os.remove")
    @patch("os.path.isfile")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_read")
    @patch("subprocess.call")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_write")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._signing_command_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.generate_random_string")
    @patch(
        "acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_signing_config_verify"
    )
    def test_057_pkcs7_sign_external(
        self,
        mock_vrf,
        mock_rand,
        mock_build,
        mock_write,
        mock_call,
        mock_read,
        mock_file,
        mock_rm,
    ):
        """test _pkcs7_sign_external() no delete"""
        mock_vrf.return_value = False
        mock_read.return_value = "foo"
        mock_call.return_value = None
        mock_file.return_value = False
        self.cahandler.signing_script_dic = {"signing_csr_path": "signing_csr_path"}
        self.assertEqual((None, "foo"), self.cahandler._pkcs7_sign_external("csr"))
        self.assertTrue(mock_rand.called)
        self.assertTrue(mock_build.called)
        self.assertTrue(mock_write.called)
        self.assertTrue(mock_call.called)
        self.assertTrue(mock_read.called)
        self.assertTrue(mock_file.called)
        self.assertFalse(mock_rm.called)

    @patch("os.remove")
    @patch("os.path.isfile")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_read")
    @patch("subprocess.call")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.binary_write")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._signing_command_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.generate_random_string")
    @patch(
        "acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_signing_config_verify"
    )
    def test_058_pkcs7_sign_external(
        self,
        mock_vrf,
        mock_rand,
        mock_build,
        mock_write,
        mock_call,
        mock_read,
        mock_file,
        mock_rm,
    ):
        """test _pkcs7_sign_external() subprocess call returns something"""
        mock_vrf.return_value = False
        mock_read.return_value = "foo"
        mock_call.return_value = 1
        mock_file.return_value = True
        self.cahandler.signing_script_dic = {"signing_csr_path": "signing_csr_path"}
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual((1, None), self.cahandler._pkcs7_sign_external("csr"))
        self.assertIn(
            "ERROR:test_a2c:Certificate enrollment aborted: 1",
            lcm.output,
        )
        self.assertTrue(mock_rand.called)
        self.assertTrue(mock_build.called)
        self.assertTrue(mock_write.called)
        self.assertTrue(mock_call.called)
        self.assertFalse(mock_read.called)
        self.assertTrue(mock_file.called)
        self.assertTrue(mock_rm.called)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._certraw_get")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._get_certificate")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_send")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_create")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._cert_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_sign_external")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_url_recode")
    def test_059_enroll(
        self,
        mock_recode,
        mock_decode,
        mock_sigext,
        mock_cert_decode,
        mock_pkcs7_cr,
        mock_encode,
        mock_sbuild,
        mock_ssend,
        mock_cert_get,
        mock_cert_raw,
    ):
        """test enroll() external signature script returning an error"""
        self.cahandler.signing_script_dic = {"foo": "bar"}
        mock_sigext.return_value = ("error", None)
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(("error", None, None, None), self.cahandler.enroll("csr"))
        self.assertIn(
            "ERROR:test_a2c:CAhandler.enroll() aborted with error: error", lcm.output
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_decode.called)
        self.assertTrue(mock_sigext.called)
        self.assertFalse(mock_cert_decode.called)
        self.assertFalse(mock_pkcs7_cr.called)
        self.assertFalse(mock_encode.called)
        self.assertFalse(mock_sbuild.called)
        self.assertFalse(mock_ssend.called)
        self.assertFalse(mock_cert_get.called)
        self.assertFalse(mock_cert_raw.called)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._certraw_get")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._get_certificate")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_send")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_create")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._cert_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_sign_external")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_url_recode")
    def test_060_enroll(
        self,
        mock_recode,
        mock_decode,
        mock_sigext,
        mock_cert_decode,
        mock_pkcs7_cr,
        mock_encode,
        mock_sbuild,
        mock_ssend,
        mock_cert_get,
        mock_cert_raw,
    ):
        """test enroll() internal signer returns error"""
        self.cahandler.signing_script_dic = {}
        mock_pkcs7_cr.return_value = ("error", None)
        mock_cert_decode.return_value = "decoded_cert"
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(("error", None, None, None), self.cahandler.enroll("csr"))
        self.assertIn(
            "ERROR:test_a2c:CAhandler.enroll() aborted with error: error", lcm.output
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_decode.called)
        self.assertFalse(mock_sigext.called)
        self.assertTrue(mock_cert_decode.called)
        self.assertTrue(mock_pkcs7_cr.called)
        self.assertFalse(mock_encode.called)
        self.assertFalse(mock_sbuild.called)
        self.assertFalse(mock_ssend.called)
        self.assertFalse(mock_cert_get.called)
        self.assertFalse(mock_cert_raw.called)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._certraw_get")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._get_certificate")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_send")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_create")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._cert_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_sign_external")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_url_recode")
    def test_061_enroll(
        self,
        mock_recode,
        mock_decode,
        mock_sigext,
        mock_cert_decode,
        mock_pkcs7_cr,
        mock_encode,
        mock_sbuild,
        mock_ssend,
        mock_cert_get,
        mock_cert_raw,
    ):
        """test enroll() - soap_request_send returns error"""
        self.cahandler.signing_script_dic = {}
        mock_cert_decode.return_value = "decoded_cert"
        mock_pkcs7_cr.return_value = (None, "pkcs_7")
        mock_ssend.return_value = ("error", None)
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual(("error", None, None, None), self.cahandler.enroll("csr"))
        self.assertIn(
            "ERROR:test_a2c:SOAP request to CA failed: error",
            lcm.output,
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_decode.called)
        self.assertFalse(mock_sigext.called)
        self.assertTrue(mock_cert_decode.called)
        self.assertTrue(mock_pkcs7_cr.called)
        self.assertTrue(mock_encode.called)
        self.assertTrue(mock_sbuild.called)
        self.assertTrue(mock_ssend.called)
        self.assertFalse(mock_cert_get.called)
        self.assertFalse(mock_cert_raw.called)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._certraw_get")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._get_certificate")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_send")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_create")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._cert_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_sign_external")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_url_recode")
    def test_062_enroll(
        self,
        mock_recode,
        mock_decode,
        mock_sigext,
        mock_cert_decode,
        mock_pkcs7_cr,
        mock_encode,
        mock_sbuild,
        mock_ssend,
        mock_cert_get,
        mock_cert_raw,
    ):
        """test enroll() - soap_request_send returns no error but no bundle"""
        self.cahandler.signing_script_dic = {}
        mock_cert_decode.return_value = "decoded_cert"
        mock_pkcs7_cr.return_value = (None, "pkcs_7")
        mock_ssend.return_value = (None, None)
        with self.assertLogs("test_a2c", level="INFO") as lcm:
            self.assertEqual((None, None, None, None), self.cahandler.enroll("csr"))
        self.assertIn(
            "ERROR:test_a2c:SOAP request to CA did not return a certificate bundle.",
            lcm.output,
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_decode.called)
        self.assertFalse(mock_sigext.called)
        self.assertTrue(mock_cert_decode.called)
        self.assertTrue(mock_pkcs7_cr.called)
        self.assertTrue(mock_encode.called)
        self.assertTrue(mock_sbuild.called)
        self.assertTrue(mock_ssend.called)
        self.assertFalse(mock_cert_get.called)
        self.assertFalse(mock_cert_raw.called)

    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._certraw_get")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._get_certificate")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_send")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._soaprequest_build")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_encode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_create")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._cert_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.CAhandler._pkcs7_sign_external")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_decode")
    @patch("acme2certifier.cahandlers.pkcs7_soap_ca_handler.b64_url_recode")
    def test_063_enroll(
        self,
        mock_recode,
        mock_decode,
        mock_sigext,
        mock_cert_decode,
        mock_pkcs7_cr,
        mock_encode,
        mock_sbuild,
        mock_ssend,
        mock_cert_get,
        mock_cert_raw,
    ):
        """test enroll() - soap_request_send returns no error but no bundle"""
        self.cahandler.signing_script_dic = {}
        mock_cert_decode.return_value = "decoded_cert"
        mock_pkcs7_cr.return_value = (None, "pkcs_7")
        mock_ssend.return_value = (None, "pkcs7_bundle")
        mock_cert_get.return_value = ["cert_1", "cert_2"]
        mock_cert_raw.return_value = "cert_raw"
        self.assertEqual(
            (None, "cert_1cert_2", "cert_raw", None), self.cahandler.enroll("csr")
        )
        self.assertTrue(mock_recode.called)
        self.assertTrue(mock_decode.called)
        self.assertFalse(mock_sigext.called)
        self.assertTrue(mock_cert_decode.called)
        self.assertTrue(mock_pkcs7_cr.called)
        self.assertTrue(mock_encode.called)
        self.assertTrue(mock_sbuild.called)
        self.assertTrue(mock_ssend.called)
        self.assertTrue(mock_cert_get.called)
        self.assertTrue(mock_cert_raw.called)


if __name__ == "__main__":

    unittest.main()



"""email handler for ACME server"""

import time
import email
import smtplib
import imaplib
import threading
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from typing import Dict, List, Callable, Optional, Any
from acme2certifier.acme_srv.helper import load_config


class EmailHandler:
    """Email handler class for sending and receiving emails"""

    def __init__(self, debug: bool = False, logger=None):
        """Initialize EmailHandler"""
        self.debug = debug
        self.logger = logger

        # IMAP configuration
        self.imap_server = None
        self.imap_port = 993
        self.imap_use_ssl = True

        # SMTP configuration
        self.smtp_server = None
        self.smtp_port = 587
        self.smtp_use_tls = True

        # Authentication
        self.username = None
        self.password = None
        self.email_address = None

        # Polling configuration
        self.polling_timer = 60  # seconds
        self.connection_timeout = 30  # seconds

        # Polling control
        self._polling_active = False
        self._polling_thread = None
        self._email_callback = None

    def __enter__(self):
        """Enter context manager"""
        self._config_load()
        return self

    def __exit__(self, *args):
        """Exit context manager"""
        self.stop_polling()

    def _config_load(self):
        """Load configuration from config file"""
        self.logger.debug("EmailHandler._config_load()")

        config_dic = load_config(self.logger, "acme_srv.cfg")

        # Load from DEFAULT section
        if "DEFAULT" in config_dic:
            # IMAP configuration
            self.imap_server = config_dic.get("DEFAULT", "imap_server", fallback=None)
            try:
                self.imap_port = int(
                    config_dic.get("DEFAULT", "imap_port", fallback=993)
                )
            except ValueError as err:
                self.logger.warning(
                    "Failed to parse imap_port from configuration. Using default 993. Error: %s",
                    err,
                )
                self.imap_port = 993

            self.imap_use_ssl = config_dic.getboolean(
                "DEFAULT", "imap_use_ssl", fallback=True
            )

            # SMTP configuration (fallback to IMAP server if not specified)
            self.smtp_server = config_dic.get(
                "DEFAULT", "smtp_server", fallback=self.imap_server
            )
            try:
                self.smtp_port = int(
                    config_dic.get("DEFAULT", "smtp_port", fallback=587)
                )
            except ValueError as err:
                self.logger.warning(
                    "Failed to parse smtp_port from configuration. Using default 587. Error: %s",
                    err,
                )
                self.smtp_port = 587

            self.smtp_use_tls = config_dic.getboolean(
                "DEFAULT", "smtp_use_tls", fallback=True
            )

            # Authentication
            self.username = config_dic.get("DEFAULT", "username", fallback=None)
            if not self.username:
                # Fallback to 'user' if 'username' is not set
                self.logger.debug("Falling back to 'user' for username")
                self.username = config_dic.get("DEFAULT", "user", fallback=None)
            self.password = config_dic.get("DEFAULT", "password", fallback=None)
            self.email_address = config_dic.get(
                "DEFAULT", "email_address", fallback=self.username
            )

            # Timing configuration
            try:
                self.polling_timer = int(
                    config_dic.get("DEFAULT", "polling_timer", fallback=60)
                )
            except ValueError as err:
                self.logger.warning(
                    "Failed to parse polling_timer from configuration. Using default 60. Error: %s",
                    err,
                )
                self.polling_timer = 60

            try:
                self.connection_timeout = int(
                    config_dic.get("DEFAULT", "connection_timeout", fallback=30)
                )
            except ValueError as err:
                self.logger.warning(
                    "Failed to parse connection_timeout from configuration. Using default 30. Error: %s",
                    err,
                )
                self.connection_timeout = 30
        else:
            self.logger.warning("DEFAULT configuration section not found")

        self.logger.debug("EmailHandler._config_load() ended")

    def send_email_challenge(self, to_address: str = None, token1: str = None):
        """send challenge email"""
        self.logger.debug("Challenge._email_send(%s)", to_address)
        message_text = f"""
This is an automatically generated ACME challenge for the email address
"{to_address}". If you did not request an S/MIME certificate for this
address, please disregard this message and consider taking appropriate
security precautions.

If you did initiate the request, your email client may be able to process
this challenge automatically. Alternatively, you may need to manually
copy the first token and paste it into the designated verification tool
or application."""

        self.send(
            to_address=to_address, subject=f"ACME: {token1}", message=message_text
        )

    def send(
        self,
        to_address: str,
        subject: str,
        message: str,
        from_address: Optional[str] = None,
        html_message: Optional[str] = None,
    ) -> bool:
        """Send email via SMTP"""
        self.logger.debug("EmailHandler.send()")

        if not self._smtp_config_validate():
            return False

        try:
            # Create message
            msg = MIMEMultipart("alternative") if html_message else MIMEText(message)
            msg["Subject"] = subject
            msg["From"] = from_address or self.email_address
            msg["To"] = to_address

            if html_message:
                # Add both plain text and HTML parts
                part1 = MIMEText(message, "plain")
                part2 = MIMEText(html_message, "html")
                msg.attach(part1)
                msg.attach(part2)

            # Connect to SMTP server
            if self.smtp_use_tls:
                server = smtplib.SMTP(
                    self.smtp_server, self.smtp_port, timeout=self.connection_timeout
                )
                server.starttls()
            else:
                server = smtplib.SMTP_SSL(
                    self.smtp_server, self.smtp_port, timeout=self.connection_timeout
                )

            # Authenticate and send
            if self.username and self.password:
                server.login(self.username, self.password)

            server.send_message(msg)
            server.quit()

            self.logger.info("Email sent successfully to %s", to_address)
            return True

        except Exception as err:
            self.logger.error("Failed to send email: %s", err)
            return False

    def receive(
        self,
        callback: Optional[Callable] = None,
        folder: str = "INBOX",
        mark_as_read: bool = True,
    ) -> List[Dict[str, Any]]:
        """Receive emails via IMAP and return as dictionary"""
        self.logger.debug("EmailHandler.receive()")

        if not self._imap_config_validate():
            return []

        try:
            mail = self._imap_connect()
            mail.login(self.username, self.password)
            mail.select(folder)

            emails = self._emails_fetch(mail, callback, mark_as_read)

            mail.close()
            mail.logout()

            self.logger.debug(
                "EmailHandler.receive(): retrieved emails: %d", bool(emails)
            )
            return emails

        except Exception as err:
            self.logger.error("Failed to receive emails: %s", err)
            return []

    def _imap_connect(self):
        """Connect to IMAP server and set timeout."""
        self.logger.debug("EmailHandler._imap_connect()")
        if self.imap_use_ssl:
            mail = imaplib.IMAP4_SSL(self.imap_server, self.imap_port)
        else:
            mail = imaplib.IMAP4(self.imap_server, self.imap_port)
        mail.socket().settimeout(self.connection_timeout)
        self.logger.debug("EmailHandler._imap_connect() ended")
        return mail

    def _emails_fetch(self, mail, callback, mark_as_read):
        """Fetch unread emails and process them."""
        self.logger.debug("EmailHandler._emails_fetch()")
        emails = []
        status, messages = mail.search(None, "UNSEEN")
        if status != "OK":
            return emails

        email_ids = messages[0].split()
        for email_id in email_ids:
            status, msg_data = mail.fetch(email_id, "(RFC822)")
            if status != "OK":
                continue
            email_body = msg_data[0][1]
            email_message = email.message_from_bytes(email_body)
            parsed_email = self._email_parse(email_message)

            if callback:
                result = callback(parsed_email)
                if result:
                    self.logger.info("Email passed filter: %s", result["subject"])
                    emails = (
                        result  # return this email only if callback returns a value
                    )
                    break
                else:
                    self.logger.debug(
                        "EmailHandler.receive(): email did not pass filter: %s",
                        parsed_email["subject"],
                    )
            else:
                emails.append(parsed_email)

            # Mark as read/unread
            if mark_as_read:
                mail.store(email_id, "+FLAGS", "\\Seen")
            else:
                mail.store(email_id, "-FLAGS", "\\Seen")
        self.logger.debug("EmailHandler._emails_fetch() ended")
        return emails

    def start_polling(
        self, callback: Callable, folder: str = "INBOX", mark_as_read: bool = True
    ):
        """Start polling for emails in a separate thread"""
        self.logger.debug("EmailHandler.start_polling()")

        if self._polling_active:
            self.logger.warning("Email polling is already active")
            return

        self._email_callback = callback
        self._polling_active = True
        self._polling_thread = threading.Thread(
            target=self._polling_loop, args=(folder, mark_as_read)
        )
        self._polling_thread.daemon = True
        self._polling_thread.start()

        self.logger.info(
            "Email polling started with %d second interval", self.polling_timer
        )

    def stop_polling(self):
        """Stop email polling"""
        self.logger.debug("EmailHandler.stop_polling()")

        if self._polling_active:
            self._polling_active = False
            if self._polling_thread:
                self._polling_thread.join(timeout=5)
            self.logger.info("Email polling stopped")

    def _polling_loop(self, folder: str, mark_as_read: bool):
        """Main polling loop (runs in separate thread)"""
        while self._polling_active:
            try:
                emails = self.receive(
                    callback=self._email_callback,
                    folder=folder,
                    mark_as_read=mark_as_read,
                )
                self.logger.debug(
                    "Polling check completed, found %d new emails", len(emails)
                )

            except Exception as err:
                self.logger.error("Error during email polling: %s", err)

            # Sleep in small increments to allow for responsive shutdown
            for _ in range(self.polling_timer):
                if not self._polling_active:
                    break
                time.sleep(1)

    def _email_parse(self, email_message) -> Dict[str, Any]:
        """Parse email message into dictionary"""
        self.logger.debug("EmailHandler._email_parse()")

        parsed = {
            "subject": email_message.get("Subject", ""),
            "from": email_message.get("From", ""),
            "to": email_message.get("To", ""),
            "date": email_message.get("Date", ""),
            "body": "",
            "html_body": "",
            "attachments": [],
        }

        # Extract body content
        if email_message.is_multipart():
            for part in email_message.walk():
                content_type = part.get_content_type()
                content_disposition = str(part.get("Content-Disposition", ""))

                if (
                    content_type == "text/plain"
                    and "attachment" not in content_disposition
                ):
                    parsed["body"] = part.get_payload(decode=True).decode(
                        "utf-8", errors="ignore"
                    )
                elif (
                    content_type == "text/html"
                    and "attachment" not in content_disposition
                ):
                    parsed["html_body"] = part.get_payload(decode=True).decode(
                        "utf-8", errors="ignore"
                    )
                elif "attachment" in content_disposition:
                    filename = part.get_filename()
                    if filename:
                        parsed["attachments"].append(
                            {
                                "filename": filename,
                                "content_type": content_type,
                                "content": part.get_payload(decode=True),
                            }
                        )
        else:
            parsed["body"] = email_message.get_payload(decode=True).decode(
                "utf-8", errors="ignore"
            )

        self.logger.debug("EmailHandler._email_parse() ended")
        return parsed

    def _smtp_config_validate(self) -> bool:
        """Validate SMTP configuration"""
        if not self.smtp_server:
            self.logger.error("SMTP server not configured")
            return False
        if not self.email_address:
            self.logger.error("Email address not configured")
            return False
        if not self.username or not self.password:
            self.logger.error("Username or password not configured")
            return False
        return True

    def _imap_config_validate(self) -> bool:
        """Validate IMAP configuration"""
        if not self.imap_server:
            self.logger.error("IMAP server not configured")
            return False
        if not self.username or not self.password:
            self.logger.error("Username or password not configured")
            return False
        return True
